#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Wed Aug 10 15:55:22 2022

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
## SHBundle original docs

# GSHS global spherical harmonic synthesis 
# f = gshs(field)

# IN:
#    field ... set of SH coefficients, either in SC-triangle or CS-square format 
#    quant ... optional string argument, defining the field quantity:
#              - 'none' ............. (default), coefficients define the output
#              - 'geoid' ............ geoid height [m],
#              - 'potential' ........ potential [m^2/s^2],
#              - 'dg', 'gravity' .... gravity anomaly [mGal], 
#              - 'tr' ............... grav. disturbance, 1st rad. derivative [mGal],
#              - 'trr' .............. 2nd rad. derivative [1/s^2],
#              - 'water' ............ equiv. water height [m],
#              - 'smd' .............. surface mass density [kg/m^2]. 
#     grd .... optional string argument, defining the grid:
#              - 'block', 'cell' .... equi-angular block midpoints. n*2n
#     n ...... grid size parameter n. (default: n = lmax, determined from field)
#              #longitude samples: 2*n
#              #latitude samples n ('blocks') or n+1.
#     h ...... (default: 0), height above Earth mean radius [m].
#     jflag .. (default: true), determines whether to subtract GRS80.

# OUT:
#    f ....... the global field
#    theRAD .. vector of co-latitudes [rad] 
#    lamRAD .. vector of longitudes [rad]

# USES:
#    vec2cs, cs2sc, eigengrav, plm, normalklm, grule, ispec
 
# SEE ALSO:
#    gsha

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

# License:
#    This file is part of PySHbundle.
#    PySHbundle is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

# Acknowledgement Statement:
#    Please note that PySHbundle has adapted the following code packages, 
#    both licensed under GNU General Public License
#       1. SHbundle: https://www.gis.uni-stuttgart.de/en/research/downloads/shbundle/

#       2. Downscaling GRACE Total Water Storage Change using Partial Least Squares Regression
#          https://springernature.figshare.com/collections/Downscaling_GRACE_Total_Water_Storage_Change_using_Partial_Least_Squares_Regression/5054564 
    
# Key Papers Referred:
#    1. Vishwakarma, B. D., Horwath, M., Devaraju, B., Groh, A., & Sneeuw, N. (2017). 
#       A data‐driven approach for repairing the hydrological catchment signal damage 
#       due to filtering of GRACE products. Water Resources Research, 
#       53(11), 9824-9844. https://doi.org/10.1002/2017WR021150

#    2. Vishwakarma, B. D., Zhang, J., & Sneeuw, N. (2021). 
#       Downscaling GRACE total water storage change using 
#       partial least squares regression. Scientific data, 8(1), 95.
#       https://doi.org/10.1038/s41597-021-00862-6

# @author: Amin Shakya, Interdisciplinary Center for Water Research (ICWaR), Indian Institute of Science (IISc)

import numpy as np
from os import chdir, getcwd

from pyshbundle.cs2sc import cs2sc
from pyshbundle.normalklm import normalklm
from pyshbundle.plm import PLM
from pyshbundle.eigengrav import eigengrav
from pyshbundle.ispec import ispec

def GSHS(field, quant = 'none', grd = 'mesh', n = -9999, h = 0, jflag = 1):
    """GSHS - Global Spherical Harmonic Synthesis

    Args:
        field (_type_): set of SH coefficients, either in SC-triangle or CS-square format
        quant (str, optional): defining the field quantity. Defaults to 'none'.
        grd (str, optional): defining the grid. Defaults to 'mesh'.
        n (int, optional): _description_. Defaults to -9999.
        h (int, optional): _description_. Defaults to 0.
        jflag (int, optional): _description_. Defaults to 1.
    
    Returns:
        f (np.ndarray): the global field
        theRAD (): vector of co-latitudes [rad]
        lamRAD (): vector of longitudes [rad]

    Raises:
        Exception: Check format of the field
        Exception: n must be scalar
        Exception: n must be integer
        Exception: Grid argument must be string
        Exception: _description_
    
    Uses:
        `cs2sc`, `normalklm`, `plm`, `eigengrav`, `ispec`

    Todo: 
        * Change general exceptions to specific and descriptive built-in ones
        + using the not and then check is not always advisable
        + Check how to document valid options
    """

    wd = getcwd()
    chdir(wd)
    
    rows, cols = field.shape
    
    if rows == cols:                    #field in CS-format 
        lmax = rows - 1
        field = cs2sc(field)
    elif cols - 2 * rows == -1:         #field in SC-format already
        lmax = rows - 1
    else:
        raise Exception("Check format of the field")
    
    if n == -9999:                      #(no value of n is input) -> set n = lmax
        n = lmax
    
    if not np.isscalar(n):
        raise Exception("n must be scalar")
    
    if not np.issubdtype(type(n), np.integer):
        raise Exception("n must be integer")
    
    if not type(grd) == str:
        raise Exception("Grid argument must be string")
        
    grd = grd.lower()
    
    
    
    #Grid Definition
    dt = np.pi/n
    
    if grd == 'pole' or grd == 'mesh':
        theRAD = np.arange(0, np.pi+dt*0.5, dt, dtype='longdouble')
        lamRAD = np.arange(0, 2*np.pi, dt, dtype='longdouble')
    elif grd == 'block' or grd == 'cell':
        theRAD = np.arange(dt/2, np.pi + dt*0.5, dt, dtype='longdouble')
        lamRAD = np.arange(dt/2, 2*np.pi + dt*0.5, dt, dtype='longdouble')
    else:
        raise Exception("Incorrect grid type input")
    
    nlat = len(theRAD)
    nlon = len(lamRAD)
    
    

#    % -------------------------------------------------------------------------
#% Preprocessing on the coefficients: 
#%    - subtract reference field (if jflag is set)
#%    - specific transfer
#%    - upward continuation
#% -------------------------------------------------------------------------
    
    if jflag:
        field = field - cs2sc(normalklm(lmax+1))
        
    l = np.arange(0, lmax+1)
    transf = np.array([eigengrav(lmax, quant, h)])[0, :, :].T
    
    field = field * np.matmul(transf, np.ones((1, 2*lmax+1)), dtype='longdouble')
    
    
    '''
    % -------------------------------------------------------------------------
% Size declarations and start the waitbar:
% Note that the definition of dlam causes straight zero-padding in case N > L.
% When N < L, there will be zero-padding up to the smallest integer multiple
% of N larger than L. After the Fourier transformation (see below), the
% proper samples have to be picked out, with stepsize dlam.
% -------------------------------------------------------------------------
    '''
    
    dlam = int(np.ceil(lmax/n))             #longitude step size
    abcols = dlam*n + 1                     #columns required in A and B
    a = np.zeros((nlat, int(abcols)), dtype='longdouble')
    b = np.zeros((nlat, int(abcols)), dtype='longdouble')
     


    m = 0
    c = field[m:lmax+1, lmax+m] 
    l = np.array([np.arange(m,lmax+1)])
    p = PLM(l, m, theRAD, nargin = 3, nargout = 1)[:,:,0]
    a[:, m] = np.dot(p,c) 
    b[:, m] = np.zeros(nlat) 
    
    
    
    for m in range(1,lmax+1,1):
        c = field[m:lmax+1,lmax+m]
        s = field[m:lmax+1,lmax-m]
        
        l = np.array([np.arange(m,lmax+1)])
        p = PLM(l, m, theRAD, nargin = 3, nargout = 1)[:,:,0]
        a[:, m] = np.dot(p,c)
        b[:, m] = np.dot(p,s)
        
    del field
    '''
     -------------------------------------------------------------------------
 The second synthesis step consists of an inverse Fourier transformation
 over the rows of a and b. 
 In case of 'block', the spectrum has to be shifted first.
 When no zero-padding has been applied, the last b-coefficients must be set to
 zero. Otherwise the number of longitude samples will be 2N+1 instead of 2N.
 For N=L this corresponds to setting SLL=0!
 -------------------------------------------------------------------------
    '''

    if grd =='block' or grd == 'cell': 
      m      = np.arange(0,abcols,1)
      cshift = np.array([np.ones(nlat)], dtype='longdouble').T * np.array([np.cos(m*np.pi/2/n)], dtype='longdouble');	# cshift/sshift describe the 
      sshift = np.array([np.ones(nlat)], dtype='longdouble').T * np.array([np.sin(m*np.pi/2/n)], dtype='longdouble');	# half-blocksize lambda shift.
      atemp  =  cshift*a + sshift*b
      b      = -sshift*a + cshift*b
      a      = atemp
    
    
    
    if np.remainder(n,lmax) == 0:               #Case without zero-padding
        b[:,abcols-1] = np.zeros(nlat)
    
    #Code for ispec
    
    f = ispec(a.T, b.T).T
    if dlam > 1: 
        f = f[:,np.arange(1,dlam*nlon+1,dlam)]

    return f, theRAD, lamRAD
    

